#ifndef CLICK_SERIALLINK2_HH
#define CLICK_SERIALLINK2_HH
#include <click/element.hh>
#include <click/timer.hh>
CLICK_DECLS

/*
 * seriallink.cc,hh
 * Eric Freudenthal (freudenthal@nyu.edu) & Tao Zhao (taozhao@cs.nyu.edu)
 * and Eddie Kohler
 *
 * Copyright (c) 2002 New York University
 *
 * For integration into the click modular router (MIT).
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, subject to the conditions
 * listed in the Click LICENSE file. These conditions include: you must
 * preserve this copyright notice, and you cannot mention the copyright
 * holders in advertising related to the Software without their permission.
 * The Software is provided WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED. This
 * notice is a summary of the Click LICENSE file; the license in that file is
 * legally binding.
 */

/*
=c
SerialLink(latency, bandwidth, queue_capacity)

=s
push-push, with scheduled events

=d
Simulation of a fixed-length queue feeding a serial 
connection with limited bandwidth & fixed end-to-end bit latency. 

Amortized constant time/packet for code in seriallink.cc/seriallink.hh.  
Priority queue in routerthread adds cost logarithmic in the number of
cocurrrent callbacks (small constants).

=h latency, read-only
Latency of single bit transferred over serial link, in seconds.

=h bandwidth, read-write
Bandwidth (in kbits/sec) of serial link.  Packet transfer time computed
as time to transfer all bits in packet + latency.

=h capacity
Capacity of queue feeding serial link, in packets.  A packet begins
transmission when the final bit of the packet preceeding it
fully enters the serial link.

Packets are dropped if queue capacity exceeded.

=a
Push, Push
*/

class SerialLink: public Element { public:
  
  SerialLink();
  ~SerialLink();
  
  const char *class_name() const	{ return "SerialLink"; }
  const char *processing() const	{ return PUSH; }
  SerialLink *clone() const		{ return new SerialLink; }
  
  int configure(Vector<String> &, ErrorHandler *);
  int initialize(ErrorHandler *);
  void cleanup(CleanupStage);
  void add_handlers();
  
  void push(int, Packet *);
  void run_timer();
  
 private:

  Packet *_qhead;
  Packet *_qtail;
  int _size;
  int _capacity;
  struct timeval _link_latency;
  uint32_t _bandwidth;		// in Kbaud
  Timer _timer;

  uint32_t _nreceived;
  uint32_t _ndrops;
  
  struct timeval serialization_delay(const Packet *) const;
  static String read_param(Element *e, void *vparam);
  static int write_handler(const String &arg, Element *e,
			   void *thunk, ErrorHandler *errh);
  
};

CLICK_ENDDECLS
#endif

